from typing import Union
import os
import numpy as np
import torch
from PIL import Image
import cv2
from .datasets import *
from .postprocess.common import *
from .types import TensorLike, ListTensorLike

LW = 2  # line width
RADIUS = 5  # circle radius
ALPHA = 0.3  # alpha for overlay
# for drawing bounding box


class Results:
    def __init__(
        self,
        pre_cfg: dict,
        post_cfg: dict,
        output: Union[TensorLike, ListTensorLike],
        **kwargs,
    ):
        self.pre_cfg = pre_cfg
        self.post_cfg = post_cfg
        self.task = post_cfg["task"]
        self.set_output(output)
        self.conf_thres = kwargs.get("conf_thres", 0.25)

    def _read_image(self, source_path: Union[str, cv2.typing.MatLike, Image.Image]):
        source_img = None

        if isinstance(source_path, Image.Image):  # PIL image open
            source_img = source_path.convert("RGB")
            source_img = np.array(source_img)
            source_img = cv2.cvtColor(source_img, cv2.COLOR_RGB2BGR)
        elif isinstance(source_path, cv2.typing.MatLike):
            source_img = np.array(
                source_path
            )  # assume imread or video read is made in BGR format
            assert (
                source_img.shape[2] == 3
            ), f"Got unexpected shape for source_img={source_img.shape}."
        else:  # str image path
            assert os.path.exists(source_path) and os.path.isfile(
                source_path
            ), f"File {source_path} does not exist or is not a file."
            source_img = cv2.imread(source_path, cv2.IMREAD_COLOR)

        return source_img

    def set_output(self, output: Union[TensorLike, ListTensorLike]):
        self.acc = None
        self.box_cls = None
        self.mask = None

        if self.task.lower() == "image_classification":
            if isinstance(output, list):
                assert len(output) == 1, f"Got unexpected output={output}."
                output = output[0]
            self.acc = output
        elif (
            self.task.lower() == "object_detection"
            or self.task.lower() == "pose_estimation"
        ):
            if isinstance(output, list):
                assert len(output) == 1, f"Got unexpected output={output}."
                output = output[0]
            self.box_cls = output
        elif self.task.lower() == "instance_segmentation":
            assert isinstance(
                output, list
            ), f"Got unexpected output={output}. It should be a list."
            if len(output) == 2:  # [box_cls, mask]
                pass
            elif len(output) == 1:  # [[box_cls, mask]]
                assert len(output[0]) == 2, f"Got unexpected output={output}."
                output = output[0]
            else:
                raise ValueError(f"Got unexpected output={output}.")
            self.box_cls = output[0]
            self.mask = output[1]
        else:
            raise NotImplementedError(
                f"Task {self.task} is not supported for plotting results."
            )
        self.output = output  # store raw output

    def plot(
        self,
        source_path: Union[str, cv2.typing.MatLike, Image.Image],
        save_path: str = None,
        **kwargs,
    ):
        if save_path is not None:
            os.makedirs(os.path.dirname(save_path), exist_ok=True)

        if self.task.lower() == "image_classification":
            return self._plot_image_classification(source_path, save_path, **kwargs)
        elif self.task.lower() == "object_detection":
            return self._plot_object_detection(source_path, save_path, **kwargs)
        elif self.task.lower() == "instance_segmentation":
            return self._plot_instance_segmentation(source_path, save_path, **kwargs)
        elif self.task.lower() == "pose_estimation":
            return self._plot_pose_estimation(source_path, save_path, **kwargs)
        else:
            raise NotImplementedError(
                f"Task {self.task} is not supported for plotting results."
            )

    def _plot_image_classification(
        self,
        source_path: Union[str, cv2.typing.MatLike, Image.Image] = None,
        save_path: str = None,
        topk=5,
        **kwargs,
    ):
        assert self.acc is not None, "No accuracy output found."
        if isinstance(self.acc, np.ndarray):
            self.acc = torch.tensor(self.acc)

        topk_probs, topk_indices = torch.topk(self.acc, topk)
        topk_probs = topk_probs.squeeze().numpy()
        topk_indices = topk_indices.squeeze().numpy()

        # load labels
        labels = [get_imagenet_label(i) for i in topk_indices]
        comments = []
        for i in range(topk):
            comments.append(f"{labels[i]}: {topk_probs[i]*100:.2f}%")
            print(f"Label: {labels[i]}, Probability: {topk_probs[i]*100:.2f}%")

        if source_path is not None and save_path is not None:
            comments = "\n".join(comments)
            img = self._read_image(source_path)
            avg_color = img.mean(axis=(0, 1))
            txt_color = (
                int(255 - avg_color[0]),
                int(255 - avg_color[1]),
                int(255 - avg_color[2]),
            )
            for i, line in enumerate(comments.splitlines()):
                (_, h), _ = cv2.getTextSize(
                    text=line,
                    fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=0.5,
                    thickness=1,
                )
                img = cv2.putText(
                    img,
                    line,
                    (15, 15 + int(1.5 * i * h)),  # line spacing
                    fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=0.5,
                    color=txt_color,
                    thickness=1,
                    lineType=cv2.LINE_AA,
                )
                cv2.imwrite(save_path, img)
                cv2.destroyAllWindows()

            return img
        else:
            return None

    def _plot_object_detection(
        self,
        source_path: Union[str, cv2.typing.MatLike, Image.Image],
        save_path: str = None,
        **kwargs,
    ):
        assert self.box_cls.shape[1] == 6 + self.post_cfg.get(
            "n_extra", 0
        ), f"Got unexpected shape for object detection box_cls={self.box_cls.shape}."

        img = self._read_image(source_path)

        self.labels = self.box_cls[:, 5].to(torch.int64)
        self.scores = self.box_cls[:, 4]
        self.boxes = scale_boxes(
            self.pre_cfg["YoloPre"]["img_size"],
            self.box_cls[:, :4],
            img.shape[:2],
        )
        contours = {i: [] for i in list(range(get_coco_class_num()))}

        for box, score, label in zip(self.boxes, self.scores, self.labels):
            img = cv2.putText(
                img,
                f"{get_coco_label(label)} {int(100*score)}%",
                (int(box[0]), int(box[1]) - 10),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.5,
                get_coco_det_palette(label),
                1,
                cv2.LINE_AA,
            )
            contours[label.item()].append(
                np.array(
                    [
                        [int(box[0]), int(box[1])],
                        [int(box[2]), int(box[1])],
                        [int(box[2]), int(box[3])],
                        [int(box[0]), int(box[3])],
                    ]
                )
            )

        for label, contour in contours.items():
            if len(contour) > 0:
                cv2.drawContours(
                    img,
                    contour,
                    -1,
                    get_coco_det_palette(label),
                    LW,
                )

        if save_path is not None:
            cv2.imwrite(save_path, img)
            cv2.destroyAllWindows()
        return img

    def _plot_instance_segmentation(
        self,
        source_path: Union[str, cv2.typing.MatLike, Image.Image],
        save_path=None,
        **kwargs,
    ):
        img = self._plot_object_detection(source_path, None, **kwargs)
        masks = scale_image(self.mask.permute(1, 2, 0), img.shape[:2])
        overlay = np.zeros((masks.shape[0], masks.shape[1], 3))

        for i, label in enumerate(self.labels):
            overlay = np.maximum(
                overlay,
                masks[:, :, i][:, :, np.newaxis]
                * np.array(get_coco_det_palette(label)).reshape(1, 1, 3),
            )

        total_mask = overlay.max(axis=2, keepdims=True)
        inv_mask = 1 - ALPHA * total_mask / 255
        img = (img * inv_mask + overlay * ALPHA).astype(np.uint8)

        if save_path is not None:
            cv2.imwrite(save_path, img)
            cv2.destroyAllWindows()
        return img

    def _plot_pose_estimation(
        self,
        source_path: Union[str, cv2.typing.MatLike, Image.Image],
        save_path=None,
        **kwargs,
    ):
        img = self._plot_object_detection(source_path, None, **kwargs)
        self.kpts = scale_coords(
            self.pre_cfg["YoloPre"]["img_size"],
            self.box_cls[:, 6:].reshape(-1, 17, 3),
            img.shape[:2],
        )
        for kpt in self.kpts:
            for i, (x, y, v) in enumerate(kpt):
                color_k = KEYPOINT_PALLETE[i]
                if v < self.conf_thres:
                    continue
                cv2.circle(
                    img,
                    (int(x), int(y)),
                    RADIUS,
                    color_k,
                    -1,
                    lineType=cv2.LINE_AA,
                )

            for j, sk in enumerate(POSE_SKELETON):
                pos1 = (int(kpt[sk[0] - 1, 0]), int(kpt[sk[0] - 1, 1]))
                pos2 = (int(kpt[sk[1] - 1, 0]), int(kpt[sk[1] - 1, 1]))

                conf1 = kpt[sk[0] - 1, 2]
                conf2 = kpt[sk[1] - 1, 2]

                if conf1 < self.conf_thres or conf2 < self.conf_thres:
                    continue
                cv2.line(
                    img,
                    pos1,
                    pos2,
                    LIMB_PALLETE[j],
                    thickness=int(np.ceil(LW / 2)),
                    lineType=cv2.LINE_AA,
                )
        if save_path is not None:
            cv2.imwrite(save_path, img)
            cv2.destroyAllWindows()
        return img



import numpy as np
import cv2
import torch
from PIL import Image
from typing import Union
from .base import PreOps
from ..types import TensorLike


class Reader(PreOps):
    def __init__(self, style: str):
        """Read image and convert to tensor"""
        assert style.lower() in [
            "pil",
            "numpy",
        ], f"Unsupported style={style} for image reader."

        self.style = style.lower()

    def __call__(self, x: Union[str, TensorLike, Image.Image]):
        if self.style == "numpy":
            if isinstance(x, np.ndarray):
                return x
            elif isinstance(x, torch.Tensor):
                return x.cpu().numpy()
            elif isinstance(x, str):
                x = cv2.imread(x)
                x = cv2.cvtColor(x, cv2.COLOR_BGR2RGB)
                return x
            elif isinstance(x, Image.Image):
                return np.array(x)
            else:
                raise ValueError("Got Unsupported Input")
        elif self.style == "pil":
            if isinstance(x, np.ndarray):
                return Image.fromarray(x.astype(np.uint8))
            elif isinstance(x, torch.Tensor):
                x = x.cpu().numpy()
                return Image.fromarray(x.astype(np.uint8))
            elif isinstance(x, str):
                return Image.open(x).convert("RGB")
            elif isinstance(x, Image.Image):
                return x
            else:
                raise ValueError("Got Unsupported Input")
        else:
            raise NotImplementedError("Got Unsupported Style")



from typing import Optional, Union, Dict, Any
import torch, os
from .._api import list_models
from ...utils import download_url_to_folder
from transformers import (
    pipeline as original_pipeline,
    AutoConfig as OriginalAutoConfig,
    AutoModel as OriginalAutoModel,
    AutoTokenizer as OriginalAutoTokenizer,
    AutoFeatureExtractor as OriginalAutoFeatureExtractor,
    AutoProcessor as OriginalAutoProcessor,
    AutoModelForSpeechSeq2Seq as OriginalAutoModelForSpeechSeq2Seq,
    AutoModelForImageTextToText as OriginalAutoModelForImageTextToText,
    AutoModelForMaskedLM as OriginalAutoModelForMaskedLM,
    AutoModelForCausalLM as OriginalAutoModelForCausalLM,
    AutoModelForVision2Seq as OriginalAutoModelForVision2Seq,
    PreTrainedModel,
    TFPreTrainedModel,
    PretrainedConfig,
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
    BaseImageProcessor,
    ProcessorMixin,
)
from transformers.pipelines.base import Pipeline
from transformers.feature_extraction_utils import PreTrainedFeatureExtractor


def convert_identifier_to_path(identifier: Any):
    if not isinstance(identifier, str):
        return identifier

    filtered = [
        model
        for models in list_models().values()
        for model in models
        if model.model_id == identifier
    ]
    model = filtered[0] if filtered else None

    if model is None:
        return identifier

    download_path = os.path.expanduser(
        f"~/.mblt_model_zoo/transformers/{model.get_directory_name()}/"
    )
    download_url_to_folder(model.download_url_base, model.file_list, download_path)

    return download_path


def pipeline(
    task: Optional[str] = None,
    model: Optional[Union[str, "PreTrainedModel", "TFPreTrainedModel"]] = None,
    config: Optional[Union[str, PretrainedConfig]] = None,
    tokenizer: Optional[
        Union[str, PreTrainedTokenizer, "PreTrainedTokenizerFast"]
    ] = None,
    feature_extractor: Optional[Union[str, PreTrainedFeatureExtractor]] = None,
    image_processor: Optional[Union[str, BaseImageProcessor]] = None,
    processor: Optional[Union[str, ProcessorMixin]] = None,
    framework: Optional[str] = None,
    revision: Optional[str] = None,
    use_fast: bool = True,
    token: Optional[Union[str, bool]] = None,
    device: Optional[Union[int, str, "torch.device"]] = None,
    device_map=None,
    torch_dtype=None,
    trust_remote_code: Optional[bool] = None,
    model_kwargs: Optional[Dict[str, Any]] = None,
    pipeline_class: Optional[Any] = None,
    **kwargs,
) -> Pipeline:
    model = convert_identifier_to_path(model)
    config = convert_identifier_to_path(config)
    tokenizer = convert_identifier_to_path(tokenizer)
    feature_extractor = convert_identifier_to_path(feature_extractor)
    image_processor = convert_identifier_to_path(image_processor)
    processor = convert_identifier_to_path(processor)

    return original_pipeline(
        task,
        model,
        config,
        tokenizer,
        feature_extractor,
        image_processor,
        processor,
        framework,
        revision,
        use_fast,
        token,
        device,
        device_map,
        torch_dtype,
        trust_remote_code,
        model_kwargs,
        pipeline_class,
        **kwargs,
    )


class AutoConfig(OriginalAutoConfig):
    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, **kwargs):
        pretrained_model_name_or_path = convert_identifier_to_path(
            pretrained_model_name_or_path
        )
        return super().from_pretrained(pretrained_model_name_or_path, **kwargs)


class AutoModel(OriginalAutoModel):
    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *model_args, **kwargs):
        pretrained_model_name_or_path = convert_identifier_to_path(
            pretrained_model_name_or_path
        )
        return super().from_pretrained(
            pretrained_model_name_or_path, *model_args, **kwargs
        )


class AutoTokenizer(OriginalAutoTokenizer):
    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *inputs, **kwargs):
        pretrained_model_name_or_path = convert_identifier_to_path(
            pretrained_model_name_or_path
        )
        return super().from_pretrained(pretrained_model_name_or_path, *inputs, **kwargs)


class AutoFeatureExtractor(OriginalAutoFeatureExtractor):
    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, **kwargs):
        pretrained_model_name_or_path = convert_identifier_to_path(
            pretrained_model_name_or_path
        )
        return super().from_pretrained(pretrained_model_name_or_path, **kwargs)


class AutoProcessor(OriginalAutoProcessor):
    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, **kwargs):
        pretrained_model_name_or_path = convert_identifier_to_path(
            pretrained_model_name_or_path
        )
        return super().from_pretrained(pretrained_model_name_or_path, **kwargs)


class AutoModelForSpeechSeq2Seq(OriginalAutoModelForSpeechSeq2Seq):
    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *model_args, **kwargs):
        pretrained_model_name_or_path = convert_identifier_to_path(
            pretrained_model_name_or_path
        )
        return super().from_pretrained(
            pretrained_model_name_or_path, *model_args, **kwargs
        )


class AutoModelForImageTextToText(OriginalAutoModelForImageTextToText):
    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *model_args, **kwargs):
        pretrained_model_name_or_path = convert_identifier_to_path(
            pretrained_model_name_or_path
        )
        return super().from_pretrained(
            pretrained_model_name_or_path, *model_args, **kwargs
        )


class AutoModelForMaskedLM(OriginalAutoModelForMaskedLM):
    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *model_args, **kwargs):
        pretrained_model_name_or_path = convert_identifier_to_path(
            pretrained_model_name_or_path
        )
        return super().from_pretrained(
            pretrained_model_name_or_path, *model_args, **kwargs
        )


class AutoModelForCausalLM(OriginalAutoModelForCausalLM):
    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *model_args, **kwargs):
        pretrained_model_name_or_path = convert_identifier_to_path(
            pretrained_model_name_or_path
        )
        return super().from_pretrained(
            pretrained_model_name_or_path, *model_args, **kwargs
        )


class AutoModelForVision2Seq(OriginalAutoModelForVision2Seq):
    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path, *model_args, **kwargs):
        pretrained_model_name_or_path = convert_identifier_to_path(
            pretrained_model_name_or_path
        )
        return super().from_pretrained(
            pretrained_model_name_or_path, *model_args, **kwargs
        )



from typing import Optional, Union, List, Tuple, Any, TypeVar

import maccel
from maccel import Cluster, Core, CoreId
import torch
import torch.nn as nn
import numpy as np

from transformers import (
    PretrainedConfig,
    AyaVisionConfig,
    AyaVisionPreTrainedModel,
    AutoModelForCausalLM,
    AutoConfig,
    AutoTokenizer,
    CohereTokenizerFast,
    AutoProcessor,
    AyaVisionProcessor,
    AutoModelForImageTextToText,
)
from transformers.models.aya_vision.modeling_aya_vision import (
    AyaVisionCausalLMOutputWithPast,
)
from transformers.utils import is_torchdynamo_compiling

from mblt_model_zoo.transformers.utils.generation_utils import MobilintGenerationMixin

from ..utils.cache_utils import MobilintCache


# type hinting: specifying the type of config class that inherits from PretrainedConfig
SpecificPretrainedConfigType = TypeVar("SpecificPretrainedConfigType", bound="PretrainedConfig")


class MobilintAyaVisionConfig(AyaVisionConfig):
    model_type = "mobilint-aya_vision"

    def __init__(
        self,
        mxq_path: str = "",
        dev_no: int = 0,
        vision_config = None,
        text_config = None,
        **kwargs,
    ):
        self.mxq_path = mxq_path
        self.dev_no = dev_no
        
        if text_config is None:
            text_config = {}
            text_config["model_type"] = "mobilint-cohere2"

        super().__init__(vision_config, text_config, **kwargs)

        if self.vision_feature_select_strategy != "full":
            raise ValueError(
                "vision_feature_select_strategy should be 'full'."
                f"Got: {self.vision_feature_select_strategy}"
            )
    
    @classmethod
    def from_dict(
        cls: type[SpecificPretrainedConfigType], config_dict: dict[str, Any], **kwargs
    ) -> SpecificPretrainedConfigType:
        return_unused_kwargs = kwargs.pop("return_unused_kwargs", False)
        
        config, kwargs = super().from_dict(config_dict, return_unused_kwargs=True, **kwargs)
        
        config.text_config.name_or_path = config.name_or_path
        config.vision_config.name_or_path = config.name_or_path
        
        if return_unused_kwargs:
            return config, kwargs
        else:
            return config


class MobilintAyaVisionForConditionalGeneration(
    AyaVisionPreTrainedModel, MobilintGenerationMixin
):
    config_class = MobilintAyaVisionConfig

    def __init__(self, config: MobilintAyaVisionConfig):
        super().__init__(config)

        self.vocab_size = config.text_config.vocab_size
        self.language_model = AutoModelForCausalLM.from_config(config.text_config)

        if self.language_model._tied_weights_keys is not None:
            self._tied_weights_keys = [
                f"language_model.{k}" for k in self.language_model._tied_weights_keys
            ]

        self.pad_token_id = (
            self.config.pad_token_id if self.config.pad_token_id is not None else -1
        )

        self.post_init()

        self.dev_no = config.dev_no
        self.acc = maccel.Accelerator(self.dev_no)
        mc = maccel.ModelConfig()
        mc.set_single_core_mode(
            core_ids=[
                CoreId(Cluster.Cluster0, Core.Core2),
                CoreId(Cluster.Cluster0, Core.Core3),
                CoreId(Cluster.Cluster1, Core.Core0),
                CoreId(Cluster.Cluster1, Core.Core1),
                CoreId(Cluster.Cluster1, Core.Core2),
                CoreId(Cluster.Cluster1, Core.Core3),
            ]
        )
        self.mxq_model = maccel.Model(f"{config.name_or_path}/{config.mxq_path}", mc)
        self.mxq_model.launch(self.acc)
    
    def get_mxq_model(self):
        return self.language_model.get_mxq_model()

    def get_input_embeddings(self):
        return self.language_model.get_input_embeddings()

    def set_input_embeddings(self, value):
        self.language_model.set_input_embeddings(value)

    def get_output_embeddings(self):
        return self.language_model.get_output_embeddings()

    def set_output_embeddings(self, new_embeddings):
        self.language_model.set_output_embeddings(new_embeddings)

    def set_decoder(self, decoder):
        self.language_model.set_decoder(decoder)

    def get_decoder(self):
        return self.language_model.get_decoder()

    def get_image_features(
        self,
        pixel_values: torch.FloatTensor,
        vision_feature_layer: Union[int, List[int]],
        vision_feature_select_strategy: str,
        **kwargs,
    ):
        if vision_feature_select_strategy != "full":
            raise ValueError(
                f"Unexpected vision_feature_select_strategy: {vision_feature_select_strategy}"
            )

        if vision_feature_layer != -1:
            raise ValueError(f"Unexpected vision_feature_layer: {vision_feature_layer}")

        image_features = pixel_values.type(torch.float32).cpu().numpy()
        image_features = self.mxq_model.infer([image_features])[0]
        image_features = np.transpose(image_features, (0, 2, 3, 1))
        image_features = torch.tensor(image_features, dtype=torch.float32, device=self.device)
        return image_features

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        pixel_values: Optional[torch.FloatTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[MobilintCache] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        vision_feature_layer: Optional[Union[int, List[int]]] = None,
        vision_feature_select_strategy: Optional[str] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        cache_position: Optional[torch.LongTensor] = None,
        logits_to_keep: Union[int, torch.Tensor] = 0,
        image_sizes: Optional[torch.Tensor] = None,
        **lm_kwargs,
    ) -> Union[Tuple, AyaVisionCausalLMOutputWithPast]:
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )
        vision_feature_layer = (
            vision_feature_layer
            if vision_feature_layer is not None
            else self.config.vision_feature_layer
        )
        vision_feature_select_strategy = (
            vision_feature_select_strategy
            if vision_feature_select_strategy is not None
            else self.config.vision_feature_select_strategy
        )

        if (input_ids is None) ^ (inputs_embeds is not None):
            raise ValueError(
                "You must specify exactly one of input_ids or inputs_embeds"
            )

        if pixel_values is not None and inputs_embeds is not None:
            raise ValueError(
                "You cannot specify both pixel_values and inputs_embeds at the same time, and must specify either one"
            )

        if inputs_embeds is None:
            inputs_embeds = self.get_input_embeddings()(input_ids)

        if pixel_values is not None:
            image_features = self.get_image_features(
                pixel_values=pixel_values,
                vision_feature_layer=vision_feature_layer,
                vision_feature_select_strategy=vision_feature_select_strategy,
                image_sizes=image_sizes,
            )

            special_image_mask = (input_ids == self.config.image_token_index).unsqueeze(
                -1
            )
            special_image_mask = special_image_mask.expand_as(inputs_embeds).to(
                inputs_embeds.device
            )
            if (
                not is_torchdynamo_compiling()
                and inputs_embeds[special_image_mask].numel() != image_features.numel()
            ):
                n_image_tokens = (input_ids == self.config.image_token_index).sum()
                n_image_features = image_features.shape[0] * image_features.shape[1]
                raise ValueError(
                    f"Image features and image tokens do not match: tokens: {n_image_tokens}, features {n_image_features}"
                )
            image_features = image_features.to(
                inputs_embeds.device, inputs_embeds.dtype
            )
            inputs_embeds = inputs_embeds.masked_scatter(
                special_image_mask, image_features
            )

        outputs = self.language_model(
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
            cache_position=cache_position,
            logits_to_keep=logits_to_keep,
            **lm_kwargs,
        )

        logits = outputs[0]

        loss = None
        if labels is not None:
            # Shift so that tokens < n predict n
            if attention_mask is not None:
                # we use the input attention mask to shift the logits and labels, because it is 2D.
                # we also crop attn mask in case it is longer, which happens in PrefixTuning with peft
                shift_attention_mask = attention_mask[:, -(logits.shape[1] - 1) :].to(
                    logits.device
                )
                shift_logits = logits[..., :-1, :][
                    shift_attention_mask.to(logits.device) != 0
                ].contiguous()
                shift_labels = labels[..., 1:][
                    shift_attention_mask.to(labels.device) != 0
                ].contiguous()
            else:
                shift_logits = logits[..., :-1, :].contiguous()
                shift_labels = labels[..., 1:].contiguous()
            # Flatten the tokens
            loss_fct = nn.CrossEntropyLoss()
            loss = loss_fct(
                shift_logits.view(-1, shift_logits.size(-1)),
                shift_labels.view(-1).to(shift_logits.device),
            )

        if not return_dict:
            output = (logits,) + outputs[1:]
            return (loss,) + output if loss is not None else output

        return AyaVisionCausalLMOutputWithPast(
            loss=loss,
            logits=logits,
            past_key_values=outputs.past_key_values,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
            image_hidden_states=image_features if pixel_values is not None else None,
        )

    def prepare_inputs_for_generation(
        self,
        input_ids,
        past_key_values=None,
        inputs_embeds=None,
        pixel_values=None,
        attention_mask=None,
        cache_position=None,
        logits_to_keep=None,
        **kwargs,
    ):
        model_inputs = self.language_model.prepare_inputs_for_generation(
            input_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            attention_mask=attention_mask,
            cache_position=cache_position,
            logits_to_keep=logits_to_keep,
            **kwargs,
        )

        if cache_position[0] == 0:
            model_inputs["pixel_values"] = pixel_values

        return model_inputs

    def tie_weights(self):
        return self.language_model.tie_weights()

    def resize_token_embeddings(
        self, new_num_tokens: Optional[int] = None, pad_to_multiple_of=None
    ) -> nn.Embedding:
        model_embeds = self.language_model.resize_token_embeddings(
            new_num_tokens, pad_to_multiple_of
        )
        # update vocab size
        self.config.text_config.vocab_size = model_embeds.num_embeddings
        self.vocab_size = model_embeds.num_embeddings
        return model_embeds

    def dispose(self):
        self.mxq_model.dispose()


AutoConfig.register("mobilint-aya_vision", MobilintAyaVisionConfig)
AutoTokenizer.register(
    MobilintAyaVisionConfig, fast_tokenizer_class=CohereTokenizerFast
)
AutoProcessor.register(MobilintAyaVisionConfig, AyaVisionProcessor)
AutoModelForImageTextToText.register(
    MobilintAyaVisionConfig, MobilintAyaVisionForConditionalGeneration
)

from ..utils.types import TransformersModelInfo

aya_vision_8b = TransformersModelInfo(
    original_model_id="CohereLabs/aya-vision-8b",
    model_id="mobilint/aya-vision-8b",
    download_url_base="https://dl.mobilint.com/model/transformers/vlm/aya-vision-8b/",
    file_list=[
        "c4ai-command-r7b-12-2024.mxq",
        "chat_template.json",
        "config.json",
        "generation_config.json",
        "model.safetensors",
        "preprocessor_config.json",
        "processor_config.json",
        "siglip2-so400m-patch14-384-vision.mxq",
        "special_tokens_map.json",
        "tokenizer.json",
        "tokenizer_config.json",
    ],
)



from typing import Union, List
import importlib
import inspect
from .wrapper import MBLT_Engine


TASKS = [
    "image_classification",
    "object_detection",
    "instance_segmentation",
    "pose_estimation",
]


def list_tasks():
    return TASKS


def list_models(tasks: Union[str, List[str]] = TASKS):
    if isinstance(tasks, str):
        tasks = [tasks]
    assert set(tasks).issubset(TASKS), f"mblt model zoo supports tasks in {TASKS}"

    available_models = {}
    for task in tasks:
        available_models[task] = []
        try:
            module = importlib.import_module(
                f".{task}", package=__name__.replace("._api", "")
            )
        except ImportError as e:
            print(f"Failed to import module for task '{task}': {e}")
            continue

        for name, obj in inspect.getmembers(module):
            if (
                inspect.isfunction(obj)
                and inspect.signature(obj).return_annotation == MBLT_Engine
            ):
                available_models[task].append(name)

    return available_models



import torch
from .yolo_anchorless_post import YOLOAnchorlessPost
from .common import *


class YOLONMSFreePost(YOLOAnchorlessPost):
    def __init__(self, pre_cfg: dict, post_cfg: dict):
        super().__init__(pre_cfg, post_cfg)
        assert (
            self.n_extra == 0
        ), "YOLOv10 is not implemented for segmentation, pose estimation"

    def decode(self, x):
        batch_box_cls = torch.cat(x, axis=-1)  # (b, no=144, 8400)

        y = []
        for xi in batch_box_cls:
            ic = (torch.amax(xi[-self.nc :, :], dim=0) > self.inv_conf_thres).to(
                self.device
            )

            xi = xi[:, ic]  # (144, *)

            if xi.numel() == 0:
                y.append(
                    torch.zeros(
                        (0, 6),
                        dtype=torch.float32,
                        device=self.device,
                    )
                )
                continue

            box, score = torch.split(
                xi[None], [self.reg_max * 4, self.nc], 1
            )  # (1, 64, *), (1, nc, *)
            dbox = (
                dist2bbox(self.dfl(box), self.anchors[:, ic], xywh=False, dim=1)
                * self.stride[:, ic]
            )
            pre_topk = (
                torch.cat([dbox, score.sigmoid()], dim=1).squeeze(0).transpose(0, 1)
            )  # (*, 84)

            max_det = min(pre_topk.shape[0], 300)

            box, score = pre_topk.split([4, self.nc], dim=-1)
            max_score = score.amax(dim=-1)
            max_score, index = torch.topk(
                max_score, max_det, dim=-1
            )  # max deteciton is 300
            index = index.unsqueeze(-1)
            box = torch.gather(box, dim=0, index=index.repeat(1, 4))
            score = torch.gather(score, dim=0, index=index.repeat(1, self.nc))

            # second topk
            score, index = torch.topk(score.flatten(), max_det)
            index = index.unsqueeze(-1)
            score = score.unsqueeze(-1)
            label = index % self.nc
            index = index // self.nc
            box = box.gather(dim=0, index=index.repeat(1, 4))

            box_cls = torch.cat([box, score, label], dim=1)  # (300, 6)
            box_cls = box_cls[box_cls[:, 4] > self.conf_thres]  # final filtering

            if box_cls.numel() == 0:
                y.append(
                    torch.zeros(
                        (0, 6),
                        dtype=torch.float32,
                        device=self.device,
                    )
                )
                continue
            y.append(box_cls)

        return y

    def nms(self, x):  # Do nothing on NMS Free model
        return x



from .base import PostBase
from .cls_post import ClsPost
from .yolo_anchor_post import YOLOAnchorPost, YOLOAnchorSegPost
from .yolo_anchorless_post import (
    YOLOAnchorlessPost,
    YOLOAnchorlessSegPost,
    YOLOAnchorlessPosePost,
)
from .yolo_nmsfree_post import YOLONMSFreePost


def build_postprocess(
    pre_cfg: dict,
    post_cfg: dict,
) -> PostBase:
    task_lower = post_cfg["task"].lower()
    if task_lower == "image_classification":
        return ClsPost(pre_cfg, post_cfg)
    elif task_lower == "object_detection":
        if post_cfg.get("anchors") is not None:
            return YOLOAnchorPost(
                pre_cfg,
                post_cfg,
            )
        elif post_cfg.get("nmsfree", False):  # nms free is only available for detection
            return YOLONMSFreePost(
                pre_cfg,
                post_cfg,
            )
        else:
            return YOLOAnchorlessPost(
                pre_cfg,
                post_cfg,
            )

    elif task_lower == "instance_segmentation":
        if post_cfg.get("anchors") is not None:
            return YOLOAnchorSegPost(
                pre_cfg,
                post_cfg,
            )
        else:
            return YOLOAnchorlessSegPost(
                pre_cfg,
                post_cfg,
            )

    elif task_lower == "pose_estimation":
        return YOLOAnchorlessPosePost(  # pose estimation is only available for anchorless model
            pre_cfg,
            post_cfg,
        )

    else:
        raise NotImplementedError(f"Task {post_cfg['task']} is not implemented yet")



from typing import List, Union
from PIL import Image
import torch
import numpy as np
import cv2
from .base import PreOps
from ..types import TensorLike


class CenterCrop(PreOps):
    def __init__(self, size: Union[int, List[int]]):
        """Center crop the image

        Args:
            size (Union[int, List[int]]): Target height and width.
        """
        super().__init__()

        if isinstance(size, list):
            assert len(size) == 2, f"Got unexpected size={size}."
            self.size = size
        elif isinstance(size, int):
            self.size = [size, size]

    def __call__(self, x: Union[TensorLike, Image.Image]):
        if isinstance(x, torch.Tensor):
            x = x.cpu().numpy()
        elif isinstance(x, Image.Image):
            x = np.array(x)

        H, W = x.shape[:2]
        if (self.size[0] == H) and (self.size[1] == W):
            return x
        elif (self.size[1] > W) or (self.size[0] > H):
            x = cv2.copyMakeBorder(
                x,
                (self.size[0] - H) // 2 if self.size[0] > H else 0,
                (self.size[0] - H + 1) // 2 if self.size[0] > H else 0,
                (self.size[1] - W) // 2 if self.size[1] > W else 0,
                (self.size[1] - W + 1) // 2 if self.size[1] > W else 0,
                cv2.BORDER_CONSTANT,
                value=0,
            )
            H, W = x.shape[:2]

        crop_top = round((H - self.size[0]) / 2.0)
        crop_left = round((W - self.size[1]) / 2.0)
        x = x[
            crop_top : crop_top + self.size[0],
            crop_left : crop_left + self.size[1],
            :,
        ]

        return x



from .yolov5_seg import YOLOv5lSeg, YOLOv5xSeg
from .yolov8_seg import YOLOv8sSeg, YOLOv8mSeg, YOLOv8lSeg, YOLOv8xSeg
from .yolov9_seg import YOLOv9cSeg
from .yolo11_seg import YOLO11lSeg, YOLO11mSeg, YOLO11xSeg



from collections import OrderedDict
from .base import PreBase
from .resize import Resize
from .center_crop import CenterCrop
from .normalize import Normalize
from .order import SetOrder
from .reader import Reader
from .yolo_pre import YoloPre


def build_preprocess(pre_cfg: OrderedDict) -> PreBase:
    res = []
    for pre_type, pre_attr in pre_cfg.items():
        pre_type_lower = pre_type.lower()
        if pre_type_lower == Reader.__name__.lower():
            res.append(Reader(**pre_attr))
        elif pre_type_lower == Resize.__name__.lower():
            res.append(Resize(**pre_attr))
        elif pre_type_lower == CenterCrop.__name__.lower():
            res.append(CenterCrop(**pre_attr))
        elif pre_type_lower == Normalize.__name__.lower():
            res.append(Normalize(**pre_attr))
        elif pre_type_lower == SetOrder.__name__.lower():
            res.append(SetOrder(**pre_attr))
        elif pre_type_lower == YoloPre.__name__.lower():
            res.append(YoloPre(**pre_attr))
        else:
            raise ValueError(f"Got unsupported pre_type={pre_type}.")

    return PreBase(res)



import os
import tempfile
import uuid
import errno
import shutil
from tqdm import tqdm
from typing import List
from urllib.request import Request, urlopen
from urllib.parse import urljoin

READ_DATA_CHUNK = 128 * 1024


# The code below is copied from torch.hub.download_url_to_file
def download_url_to_file(
    url: str,
    dst: str,
    progress: bool = True,
) -> None:
    r"""Download object at the given URL to a local path.

    Args:
        url (str): URL of the object to download
        dst (str): Full path where object will be saved, e.g. ``/tmp/temporary_file``
        progress (bool, optional): whether or not to display a progress bar to stderr
            Default: True
    """
    file_size = None
    req = Request(url, headers={"User-Agent": "mblt_model_zoo"})
    u = urlopen(req)
    meta = u.info()
    if hasattr(meta, "getheaders"):
        content_length = meta.getheaders("Content-Length")
    else:
        content_length = meta.get_all("Content-Length")
    if content_length is not None and len(content_length) > 0:
        file_size = int(content_length[0])

    # We deliberately save it in a temp file and move it after
    # download is complete. This prevents a local working checkpoint
    # being overridden by a broken download.
    # We deliberately do not use NamedTemporaryFile to avoid restrictive
    # file permissions being applied to the downloaded file.
    dst = os.path.expanduser(dst)
    for seq in range(tempfile.TMP_MAX):
        tmp_dst = dst + "." + uuid.uuid4().hex + ".partial"
        try:
            f = open(tmp_dst, "w+b")
        except FileExistsError:
            continue
        break
    else:
        raise FileExistsError(errno.EEXIST, "No usable temporary file name found")

    try:
        with tqdm(
            desc=os.path.split(dst)[1],
            total=file_size,
            disable=not progress,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        ) as pbar:
            while True:
                buffer = u.read(READ_DATA_CHUNK)
                if len(buffer) == 0:
                    break
                f.write(buffer)  # type: ignore[possibly-undefined]
                pbar.update(len(buffer))

        f.close()
        shutil.move(f.name, dst)
    finally:
        f.close()
        if os.path.exists(f.name):
            os.remove(f.name)


def download_url_to_folder(
    url_dir: str,
    url_filelist: List[str],
    dst: str,
    progress: bool = True,
) -> None:
    os.makedirs(dst, exist_ok=True)

    for url_file in url_filelist:
        u = urljoin(url_dir, url_file)
        d = os.path.join(dst, url_file)
        try:
            if not os.path.exists(d):
                download_url_to_file(u, d, progress)

        except Exception:
            print(f"{url_file} download failed")



import torch
import numpy as np
from typing import Union
from .base import PostBase
from ..types import TensorLike, ListTensorLike


class ClsPost(PostBase):
    def __init__(self, pre_cfg: dict, post_cfg: dict):
        super().__init__()

    def __call__(self, x: Union[TensorLike, ListTensorLike]):
        if isinstance(x, list):
            assert (
                len(x) == 1
            ), "assume that classification model only returns pre-softmax tensor"
            x = x[0]

        if isinstance(x, np.ndarray):
            x = torch.from_numpy(x).to(self.device)
        else:
            x = x.to(self.device)

        if x.ndim == 3:
            x = x.unsqueeze(0)
        assert (
            x.ndim == 4
        ), f"Assume that the result is always in form of NCHW. But the shape is {x.shape}"

        x = x.flatten(1)  # assume that the shape can be made to (b, 1000)
        return x.softmax(dim=-1)



from typing import List, Union, Tuple
import numpy as np
import torch
from PIL import Image
import torch.nn.functional as F
from .base import PreOps
from ..types import TensorLike

PIL_INTERP_CODES = {
    "nearest": Image.Resampling.NEAREST,
    "bilinear": Image.Resampling.BILINEAR,
    "bicubic": Image.Resampling.BICUBIC,
    "box": Image.Resampling.BOX,
    "hamming": Image.Resampling.HAMMING,
    "lanczos": Image.Resampling.LANCZOS,
}


class Resize(PreOps):
    """Resize image in Torch backend"""

    def __init__(
        self,
        size: Union[int, List[int]],
        interpolation: str,
    ):
        # Note that this behaves different for npy image and PIL image
        super().__init__()
        self.size = size  # h, w
        self.interpolation = interpolation

    def __call__(self, x: Union[TensorLike, Image.Image]):
        """Resize image

        Args:
            x (Union[np.ndarray, Image]): Image to be resized.

        Raises:
            TypeError: If x is not numpy array or PIL image.

        Returns:
            x: Resized image.
        """
        # result: np.ndarray (H, W, C)
        if isinstance(x, np.ndarray):
            x = torch.from_numpy(x).to(self.device)

        if isinstance(x, torch.Tensor):
            assert x.ndim() == 3, f"Got unexpected x.shape={x.shape}."
            x = x.to(self.device)
            img_h, img_w = x.shape[:2]
            new_h, new_w = self._compute_resized_output_size(img_h, img_w)

            if [img_h, img_w] == [new_h, new_w]:
                return x

            x, need_cast, need_squeeze, out_dtype = self._cast_squeeze_in(
                x, [torch.float32, torch.float64]
            )

            x = F.interpolate(
                x[None],
                size=(new_h, new_w),
                mode=self.interpolation,
                align_corners=(
                    False if self.interpolation in ["bilinear", "bicubic"] else None
                ),
                antialias=self.interpolation in ["bilinear", "bicubic"],
            )

            x = self._cast_squeeze_out(x, need_cast, need_squeeze, out_dtype)

            return x.to(self.device)
        elif isinstance(x, Image.Image):
            img_w, img_h = x.size
            new_h, new_w = self._compute_resized_output_size(img_h, img_w)
            if [img_h, img_w] == [new_h, new_w]:
                return x
            return x.resize(
                size=(new_w, new_h),
                resample=PIL_INTERP_CODES[self.interpolation],
            )
        else:
            raise TypeError(f"Got unexpected type for x={type(x)}.")

    def _compute_resized_output_size(self, img_h: int, img_w: int):
        if isinstance(self.size, int):
            # to match the shortest side to self.size with the same ratio
            ratio = max(self.size / img_h, self.size / img_w)
            new_h = int(round(img_h * ratio))
            new_w = int(round(img_w * ratio))
        elif isinstance(self.size, list) and len(self.size) == 2:
            new_h, new_w = self.size
        else:
            raise ValueError(f"Got unexpected size={self.size}.")

        return [new_h, new_w]

    def _cast_squeeze_in(
        self, img: torch.Tensor, req_dtypes: List[torch.dtype]
    ) -> Tuple[torch.Tensor, bool, bool, torch.dtype]:
        need_squeeze = False
        # make image NCHW
        if img.ndim < 4:
            img = img.unsqueeze(dim=0)
            need_squeeze = True

        out_dtype = img.dtype
        need_cast = False
        if out_dtype not in req_dtypes:
            need_cast = True
            req_dtype = req_dtypes[0]
            img = img.to(req_dtype)
        return img, need_cast, need_squeeze, out_dtype

    def _cast_squeeze_out(
        self,
        img: torch.Tensor,
        need_cast: bool,
        need_squeeze: bool,
        out_dtype: torch.dtype,
    ) -> torch.Tensor:
        if need_squeeze:
            img = img.squeeze(dim=0)

        if need_cast:
            if out_dtype in (
                torch.uint8,
                torch.int8,
                torch.int16,
                torch.int32,
                torch.int64,
            ):
                # it is better to round before cast
                img = torch.round(img)
            img = img.to(out_dtype)

        return img



import numpy as np
import torch
import torch.nn.functional as F
import cv2
from typing import Union


def xywh2xyxy(x: Union[np.ndarray, torch.Tensor]):
    # Convert bounding box coordinates from (cx, cy, width, height) format to (x1, y1, x2, y2) format where (x1, y1) is the top-left corner and (x2, y2) is the bottom-right corner.
    if isinstance(x, np.ndarray):
        y = np.copy(x)
    elif isinstance(x, torch.Tensor):
        y = torch.clone(x).to(x)
    else:
        raise ValueError("x should be np.ndarray or torch.Tensor")

    y[..., 0] = x[..., 0] - x[..., 2] / 2
    y[..., 1] = x[..., 1] - x[..., 3] / 2
    y[..., 2] = x[..., 0] + x[..., 2] / 2
    y[..., 3] = x[..., 1] + x[..., 3] / 2

    return y


def xyxy2xywh(x: Union[np.ndarray, torch.Tensor]):
    # Convert bounding box coordinates from (x1, y1, x2, y2) format to (cx, cy, width, height) format where (cx, cy) is the center of the bounding box and width and height are the dimensions of the bounding box.
    if isinstance(x, np.ndarray):
        y = np.copy(x)
    elif isinstance(x, torch.Tensor):
        y = torch.clone(x).to(x)
    else:
        raise ValueError("x should be np.ndarray or torch.Tensor")

    y[..., 0] = (x[..., 0] + x[..., 2]) / 2
    y[..., 1] = (x[..., 1] + x[..., 3]) / 2
    y[..., 2] = x[..., 2] - x[..., 0]
    y[..., 3] = x[..., 3] - x[..., 1]

    return y


def compute_ratio_pad(img1_shape, img0_shape, ratio_pad=None):
    """Compute ratio and pad which were used to resize image to input_shape"""
    if ratio_pad is None:  # calculate from img0_shape
        gain = min(
            img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1]
        )  # gain  = old / new
        pad = round((img1_shape[1] - img0_shape[1] * gain) / 2 - 0.1), round(
            (img1_shape[0] - img0_shape[0] * gain) / 2 - 0.1
        )  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]
    return gain, pad


def dist2bbox(distance, anchor_points, xywh=True, dim=-1):
    """Transform distance(ltrb) to box(xywh or xyxy)."""
    lt, rb = distance.chunk(2, dim)
    x1y1 = anchor_points - lt
    x2y2 = anchor_points + rb
    if xywh:
        return torch.cat(((x1y1 + x2y2) / 2, x2y2 - x1y1), dim)  # xywh bbox
    else:
        return torch.cat((x1y1, x2y2), dim)  # xyxy bbox


def clip_boxes(boxes, shape):
    boxes[..., 0] = boxes[..., 0].clamp(0, shape[1])
    boxes[..., 1] = boxes[..., 1].clamp(0, shape[0])
    boxes[..., 2] = boxes[..., 2].clamp(0, shape[1])
    boxes[..., 3] = boxes[..., 3].clamp(0, shape[0])
    return boxes


def clip_coords(coords, shape):
    """Clip coordinates to image shape."""
    coords[..., 0] = coords[..., 0].clamp(0, shape[1])
    coords[..., 1] = coords[..., 1].clamp(0, shape[0])
    return coords


def scale_boxes(img1_shape, boxes, img0_shape, ratio_pad=None, padding=True):
    """
    Original Source: https://github.com/ultralytics/ultralytics/blob/main/ultralytics/utils/ops.py#L92
    Rescales bounding boxes (in the format of xyxy) from the shape of the image they were originally specified in
    (img1_shape) to the shape of a different image (img0_shape).

    Args:
        img1_shape (tuple): The shape of the image that the bounding boxes are for, in the format of (height, width).
        boxes (np.ndarray): the bounding boxes of the objects in the image, in the format of (x1, y1, x2, y2)
        img0_shape (tuple): the shape of the target image, in the format of (height, width).
        ratio_pad (tuple): a tuple of (ratio, pad) for scaling the boxes. If not provided, the ratio and pad will be
            calculated based on the size difference between the two images.
        padding (bool): If True, assuming the boxes is based on image augmented by yolo style. If False then do regular
            rescaling.

    Returns:
        boxes (np.ndarray): The scaled bounding boxes, in the format of (x1, y1, x2, y2)
    """
    gain, pad = compute_ratio_pad(img1_shape, img0_shape, ratio_pad)

    if padding:
        boxes[..., [0, 2]] -= pad[0]  # x padding
        boxes[..., [1, 3]] -= pad[1]  # y padding
    boxes[..., :4] /= gain
    return clip_boxes(boxes, img0_shape)


def scale_coords(img1_shape, coords, img0_shape, ratio_pad=None, padding=True):
    """
    Original Source:
    https://github.com/ultralytics/ultralytics/blob/main/ultralytics/utils/ops.py#L756

    Args:
        img1_shape (tuple): The shape of the image that the bounding boxes are for, in the format of (height, width).
        coords (tuple): The coordinates of the objects in the image, in the format of (x, y).
        img0_shape (tuple): The shape of the target image, in the format of (height, width).
        ratio_pad (tuple): a tuple of (ratio, pad) for scaling the boxes. If not provided, the ratio and pad will be
            calculated based on the size difference between the two images.
        padding (bool): If True, assuming the boxes is based on image augmented by yolo style. If False then do regular
            rescaling.

    Returns:
        coords (tuple): The scaled coordinates, in the format of (x, y)
    """

    if ratio_pad is None:  # calculate from img0_shape
        gain = min(
            img1_shape[0] / img0_shape[0], img1_shape[1] / img0_shape[1]
        )  # gain  = old / new
        pad = int((img1_shape[1] - img0_shape[1] * gain) / 2), int(
            (img1_shape[0] - img0_shape[0] * gain) / 2
        )  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]

    if padding:
        coords[..., 0] -= pad[0]  # x padding
        coords[..., 1] -= pad[1]  # y padding

    coords[..., :2] /= gain
    return clip_coords(coords, img0_shape)


def process_mask(protos, masks_in, bboxes, shape, upsample=False):
    """
    https://github.com/ultralytics/ultralytics/blob/main/ultralytics/utils/ops.py#L680
    Faster way to process masks.
    protos: [mask_dim, mask_h, mask_w]
    masks_in: [n, mask_dim], n is number of masks after nms
    bboxes: [n, 4], n is number of masks after nms
    shape: input_image_size, (h, w)

    return: h, w, n
    """

    c, mh, mw = protos.shape  # CHW
    ih, iw = shape
    masks = (masks_in @ protos.float().view(c, -1)).view(-1, mh, mw)  # n, CHW

    downsampled_bboxes = bboxes.clone()
    downsampled_bboxes[:, 0] *= mw / iw
    downsampled_bboxes[:, 2] *= mw / iw
    downsampled_bboxes[:, 3] *= mh / ih
    downsampled_bboxes[:, 1] *= mh / ih

    masks = crop_mask(masks, downsampled_bboxes)  # CHW

    if upsample:
        masks = F.interpolate(masks[None], shape, mode="bilinear", align_corners=False)[
            0
        ]  # CHW
    return masks.gt_(0.0)


def process_mask_upsample(protos, masks_in, bboxes, shape):
    """
    Takes the output of the mask head, and applies the mask to the bounding boxes. This produces masks of higher quality
    but is slower.

    https://github.com/ultralytics/ultralytics/blob/main/ultralytics/utils/ops.py#L713
    Args:
        protos (torch.Tensor): [mask_dim, mask_h, mask_w]
        masks_in (torch.Tensor): [n, mask_dim], n is number of masks after nms
        bboxes (torch.Tensor): [n, 4], n is number of masks after nms
        shape (tuple): the size of the input image (h,w)

    Returns:
        (torch.Tensor): The upsampled masks.
    """
    c, mh, mw = protos.shape  # CHW
    masks = (masks_in @ protos.float().view(c, -1)).view(-1, mh, mw)  # n, CHW
    masks = scale_masks(masks[None], shape)[0]  # CHW
    masks = crop_mask(masks, bboxes)  # CHW
    return masks.gt_(0.0)


def crop_mask(masks, boxes):
    """
    Crop masks to bounding boxes.

    Args:
        masks (torch.Tensor): [n, h, w] tensor of masks.
        boxes (torch.Tensor): [n, 4] tensor of bbox coordinates in relative point form.

    Returns:
        (torch.Tensor): Cropped masks.
    """
    _, h, w = masks.shape
    x1, y1, x2, y2 = torch.chunk(boxes[:, :, None], 4, 1)  # x1 shape(n,1,1)
    r = torch.arange(w, device=masks.device, dtype=x1.dtype)[
        None, None, :
    ]  # rows shape(1,1,w)
    c = torch.arange(h, device=masks.device, dtype=x1.dtype)[
        None, :, None
    ]  # cols shape(1,h,1)

    return masks * ((r >= x1) * (r < x2) * (c >= y1) * (c < y2))


def scale_masks(masks, shape, padding=True):
    """
    Rescale segment masks to shape.

    Args:
        masks (torch.Tensor): (N, C, H, W).
        shape (tuple): Height and width.
        padding (bool): If True, assuming the boxes is based on image augmented by yolo style. If False then do regular
            rescaling.

    Returns:
        (torch.Tensor): Rescaled masks.
    """
    mh, mw = masks.shape[2:]
    gain = min(mh / shape[0], mw / shape[1])  # gain  = old / new
    pad = [mw - shape[1] * gain, mh - shape[0] * gain]  # wh padding
    if padding:
        pad[0] /= 2
        pad[1] /= 2
    top, left = (int(pad[1]), int(pad[0])) if padding else (0, 0)  # y, x
    bottom, right = (int(mh - pad[1]), int(mw - pad[0]))
    masks = masks[..., top:bottom, left:right]

    masks = F.interpolate(masks, shape, mode="bilinear", align_corners=False)  # NCHW
    return masks


def scale_image(masks, im0_shape, ratio_pad=None):
    """
    Takes a mask, and resizes it to the original image size.
    Original Source:
        https://github.com/ultralytics/ultralytics/blob/main/ultralytics/utils/ops.py#L377
    Args:
        masks (np.ndarray): resized and padded masks/images, [h, w, num]/[h, w, 3].
        im0_shape (tuple): the original image shape
        ratio_pad (tuple): the ratio of the padding to the original image.

    Returns:
        masks (np.ndarray): The masks that are being returned.
    """
    # Rescale coordinates (xyxy) from im1_shape to im0_shape
    if isinstance(masks, torch.Tensor):
        masks = masks.cpu().numpy()

    im1_shape = masks.shape
    if im1_shape[:2] == im0_shape[:2]:
        return masks
    elif im1_shape[2] == 0:
        return np.zeros((im0_shape[0], im0_shape[1], 0), dtype=np.float32)

    if ratio_pad is None:  # calculate from im0_shape
        gain = min(
            im1_shape[0] / im0_shape[0], im1_shape[1] / im0_shape[1]
        )  # gain  = old / new
        pad = (im1_shape[1] - im0_shape[1] * gain) / 2, (
            im1_shape[0] - im0_shape[0] * gain
        ) / 2  # wh padding
    else:
        gain = ratio_pad[0][0]
        pad = ratio_pad[1]
    top, left = int(pad[1]), int(pad[0])  # y, x
    bottom, right = int(im1_shape[0] - pad[1]), int(im1_shape[1] - pad[0])

    if len(masks.shape) < 2:
        raise ValueError(
            f'"len of masks shape" should be 2 or 3, but got {len(masks.shape)}'
        )
    masks = masks[top:bottom, left:right]
    masks = cv2.resize(masks, (im0_shape[1], im0_shape[0]))
    if len(masks.shape) == 2:
        masks = masks[:, :, None]

    return masks


def non_max_suppression(boxes, scores, iou_threshold, max_output):
    """
    Modified non-maximum suppression (NMS) implemented with PyTorch.

    Args:
        boxes (torch.Tensor): Bounding boxes in (x1, y1, x2, y2) format.
        scores (torch.Tensor): Confidence scores for each box (assumed to be sorted in descending order).
        iou_threshold (float): IoU threshold for suppression.
        max_output (int): Maximum number of boxes to keep.

    Returns:
        List[int]: Indices of the boxes that have been kept after NMS.
    """
    if boxes.numel() == 0:
        return []

    # Coordinates of bounding boxes
    start_x = boxes[:, 0]
    start_y = boxes[:, 1]
    end_x = boxes[:, 2]
    end_y = boxes[:, 3]

    picked_indices = []

    # Compute areas of bounding boxes
    areas = (end_x - start_x) * (end_y - start_y)

    # Create an index order (assumed scores are already sorted in descending order)
    order = torch.arange(scores.size(0)).to(boxes.device)

    while order.numel() > 0 and len(picked_indices) < max_output:
        # The index with the highest score
        index = order[0].item()
        picked_indices.append(index)
        order = order[1:]  # Remove the index from the order

        if order.numel() == 0 or len(picked_indices) >= max_output:
            break

        # Compute the coordinates of the intersection boxes
        x1 = torch.maximum(start_x[index], start_x[order])
        y1 = torch.maximum(start_y[index], start_y[order])
        x2 = torch.minimum(end_x[index], end_x[order])
        y2 = torch.minimum(end_y[index], end_y[order])

        # Compute width and height of the intersection boxes
        w = torch.clamp(x2 - x1, min=0.0)
        h = torch.clamp(y2 - y1, min=0.0)
        intersection = w * h

        # Compute the IoU ratio
        union = areas[index] + areas[order] - intersection
        ratio = intersection / union

        # Keep boxes with IoU less than or equal to the threshold
        keep = (ratio <= iou_threshold).to(order.device)
        order = order[keep]

    return picked_indices



from typing import Optional, Union, Any, TypeVar

import maccel
import os
import torch
import torch.nn as nn
import numpy as np
import math
from PIL import Image
from cv2 import resize as cv2_resize, INTER_CUBIC
import torch.nn.functional as F

from transformers import (
    Qwen2VLTextConfig,
    Qwen2VLConfig,
    Qwen2TokenizerFast,
    Qwen2VLProcessor,
    Qwen2VLPreTrainedModel,
    Qwen2VLModel,
    Qwen2VLForConditionalGeneration,

    PretrainedConfig,
    AutoConfig,
    AutoTokenizer,
    AutoProcessor,
    AutoModelForImageTextToText,
)
from transformers.models.qwen2_vl.configuration_qwen2_vl import Qwen2VLVisionConfig
from transformers.utils import logging
from transformers.modeling_outputs import BaseModelOutputWithPast
from transformers.image_utils import ImageInput
from transformers.video_utils import VideoInput
from transformers.processing_utils import Unpack
from transformers.tokenization_utils_base import PreTokenizedInput, TextInput
from transformers.feature_extraction_utils import BatchFeature
from transformers.models.qwen2_vl.processing_qwen2_vl import Qwen2VLProcessorKwargs
from transformers.image_utils import load_image

from mblt_model_zoo.transformers.utils.generation_utils import MobilintGenerationMixin

from ..utils.cache_utils import MobilintCache


logger = logging.get_logger(__name__)


# type hinting: specifying the type of config class that inherits from PretrainedConfig
SpecificPretrainedConfigType = TypeVar("SpecificPretrainedConfigType", bound="PretrainedConfig")


class MobilintQwen2VLProcessor(Qwen2VLProcessor):
    def __call__(
        self,
        images: ImageInput = None,
        text: Union[TextInput, PreTokenizedInput, list[TextInput], list[PreTokenizedInput]] = None,
        videos: VideoInput = None,
        **kwargs: Unpack[Qwen2VLProcessorKwargs],
    ) -> BatchFeature:
        # Make sure images is only one instance of PIL.Image.Image, np.ndarray, torch.Tensor, or None
        while isinstance(images, list):
            if len(images) > 1:
                raise NotImplementedError("Only one image input is supported")
            images = images[0]
        
        if isinstance(images, str):
            images = load_image(images)
        
        # Image should be resized into (224, 224) to fit image token position
        size = (224, 224)
        
        if isinstance(images, Image.Image):
            images = images.resize(size)
        elif isinstance(images, np.ndarray):
            if images.ndim == 2:  # 흑백
                return cv2_resize(images, size[::-1], interpolation=INTER_CUBIC)
            elif images.ndim == 3:
                return cv2_resize(images, size[::-1], interpolation=INTER_CUBIC)
            else:
                raise ValueError(f"Unsupported ndarray shape: {images.shape}")
        elif torch.is_tensor(images):
            if images.ndim == 3:  # CHW
                images = images.unsqueeze(0).float()  # BCHW
                images = F.interpolate(images, size=size, mode="bicubic", align_corners=False)
            elif images.ndim == 2:  # HW
                images = images.unsqueeze(0).unsqueeze(0).float()  # B1HW
                images = F.interpolate(images, size=size, mode="bicubic", align_corners=False)
            else:
                raise ValueError(f"Unsupported tensor shape: {tuple(images.shape)}")
        else:
            raise TypeError(f"Unsupported type of image: {type(images)}")
        
        if videos is not None:
            raise NotImplementedError("Video inputs are not supported")
                    
        return super().__call__(images, text, videos, **kwargs)

class MobilintQwen2VLTextConfig(Qwen2VLTextConfig):
    model_type = "mobilint-qwen2_vl_text"
    keys_to_ignore_at_inference = []
    base_model_tp_plan = {}
    base_model_pp_plan = {}

    def __init__(
        self,
        mxq_path: str = "",
        dev_no: int = 0,
        **kwargs,
    ):
        self.mxq_path = mxq_path
        self.dev_no = dev_no

        super().__init__(**kwargs)

        self.tie_word_embeddings = False


class MobilintQwen2VLVisionConfig(Qwen2VLVisionConfig):
    model_type = "mobilint-qwen2_vl"
    
    def __init__(
        self,
        mxq_path: str = "",
        dev_no: int = 0,
        **kwargs,
    ):
        self.mxq_path = mxq_path
        self.dev_no = dev_no

        super().__init__(**kwargs)
        

class MobilintQwen2VLConfig(Qwen2VLConfig):
    model_type = "mobilint-qwen2_vl"
    sub_configs = {"vision_config": MobilintQwen2VLVisionConfig, "text_config": MobilintQwen2VLTextConfig}
    keys_to_ignore_at_inference = []
    
    @classmethod
    def from_dict(
        cls: type[SpecificPretrainedConfigType], config_dict: dict[str, Any], **kwargs
    ) -> SpecificPretrainedConfigType:
        return_unused_kwargs = kwargs.pop("return_unused_kwargs", False)
        
        config, kwargs = super().from_dict(config_dict, return_unused_kwargs=True, **kwargs)
        
        config.text_config.name_or_path = config.name_or_path
        config.vision_config.name_or_path = config.name_or_path
        
        if return_unused_kwargs:
            return config, kwargs
        else:
            return config
        
        
class MobilintQwen2VLPreTrainedModel(Qwen2VLPreTrainedModel):
    config: MobilintQwen2VLConfig
    supports_gradient_checkpointing = False
    _no_split_modules = []
    _supports_flash_attn = False
    _supports_sdpa = False

    _can_compile_fullgraph = False
    _supports_attention_backend = False


class MobilintQwen2VisionTransformerPretrainedModel(MobilintQwen2VLPreTrainedModel):
    config: MobilintQwen2VLVisionConfig
    _no_split_modules = []
    
    def __init__(self, config: MobilintQwen2VLVisionConfig) -> None:
        super().__init__(config)
        self.spatial_merge_size = config.spatial_merge_size

        self.gradient_checkpointing = False
                
        self.dev_no = config.dev_no
        self.acc = maccel.Accelerator(self.dev_no)
        mc = maccel.ModelConfig()
        mc.set_multi_core_mode([maccel.Cluster.Cluster1])
        self.mxq_model = maccel.Model(f"{config.name_or_path}/{config.mxq_path}", mc)
        self.mxq_model.launch(self.acc)
    
    def __getattribute__(self, name: str, /) -> Any:
        if name == 'dtype':
            return self.get_dtype()
        elif name == 'device':
            return self.get_device()
        else:
            return super().__getattribute__(name)
    
    def get_dtype(self) -> torch.dtype:
        return torch.float32

    def get_device(self) -> torch.device:
        return 'cpu'

    def forward(
        self,
        hidden_states: torch.Tensor,
        grid_thw: torch.Tensor,
    ) -> torch.Tensor:
        gt, gh, gw = grid_thw[0].tolist()

        c  = 3
        pt = 2
        Mh = 2
        Mw = 2
        gh2 = gh // 2
        gw2 = gw // 2
        ph = pw = int((hidden_states.shape[-1] // (pt * c)) ** 0.5)

        assert hidden_states.shape[0] == gt * gh2 * gw2 * Mh * Mw
        assert hidden_states.shape[1] == c * pt * ph * pw

        # (gt, gh2, gw2, Mh, Mw, c, pt, ph, pw)
        hidden_states = hidden_states.view(gt, gh2, gw2, Mh, Mw, c, pt, ph, pw)

        # rearrange: "(gt gh gw Mh Mw) (c pt ph pw) -> gt (gh gw ph) (Mh Mw pw) (pt c)"
        # (gt, pt, c, Mh, Mw, pw, gh2, gw2, ph)
        hidden_states = hidden_states.permute(0, 1, 2, 7, 3, 4, 8, 6, 5).contiguous()

        # gt (gh gw ph) (Mh Mw pw) (pt c)
        hidden_states = hidden_states.view(
            gt,
            gh2 * gw2 * ph,
            Mh * Mw * pw,
            pt * c,
        ).squeeze(0)

        hidden_states = hidden_states.to(torch.float32).cpu().numpy()
        image_embeds = self.mxq_model.infer(hidden_states)[0]
        image_embeds = torch.tensor(image_embeds, dtype=torch.float32, device=self.device).squeeze(0).squeeze(0)
        
        return image_embeds

    def dispose(self):
        self.mxq_model.dispose()


class MobilintQwen2VLTextModel(MobilintQwen2VLPreTrainedModel):
    config: MobilintQwen2VLTextConfig
    
    def __init__(self, config: MobilintQwen2VLTextConfig):
        super().__init__(config)
        self.padding_idx = config.pad_token_id
        self.vocab_size = config.vocab_size

        self.embed_tokens = nn.Embedding(config.vocab_size, config.hidden_size, self.padding_idx)
        self.has_sliding_layers = "sliding_attention" in self.config.layer_types

        self.gradient_checkpointing = False
        
        self.dev_no = config.dev_no
        self.acc = maccel.Accelerator(self.dev_no)
        mc = maccel.ModelConfig()
        mc.set_single_core_mode(1)
        self.mxq_model = maccel.Model(f"{config.name_or_path}/{config.mxq_path}", mc)
        self.mxq_model.launch(self.acc)
    
    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[MobilintCache] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        cache_position: Optional[torch.LongTensor] = None,
        chunk_size: int = 0,
    ) -> Union[tuple, BaseModelOutputWithPast]:
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache

        return_dict = return_dict if return_dict is not None else self.config.use_return_dict
        
        if output_attentions:
            logger.warning_once("output_attentions is not supported.")

        if output_hidden_states:
            logger.warning_once("output_hidden_states is not supported.")

        if (input_ids is None) ^ (inputs_embeds is not None):
            raise ValueError("You must specify exactly one of input_ids or inputs_embeds")

        if self.gradient_checkpointing and self.training:
            if use_cache:
                logger.warning_once(
                    "`use_cache=True` is incompatible with gradient checkpointing. Setting `use_cache=False`..."
                )
                use_cache = False

        if inputs_embeds is None:
            inputs_embeds = self.embed_tokens(input_ids)

        if cache_position is None:
            past_seen_tokens = past_key_values.get_seq_length() if past_key_values is not None else 0
            cache_position = torch.arange(
                past_seen_tokens, past_seen_tokens + inputs_embeds.shape[1], device=inputs_embeds.device
            )

        inputs_embeds = inputs_embeds.type(torch.float32).cpu().numpy()

        if inputs_embeds.ndim == 3:
            inputs_embeds = np.expand_dims(
                inputs_embeds, 1
            )  # (batch, 1, seqlen, hidden_size)

        # max width should be appropriate number for chunking (ex. 192 for Llama 3.2 3B)
        # it should be searched experimentally
        if chunk_size == 0:
            chunk_size = self.mxq_model.get_input_buffer_info()[0].max_width
        num_of_chunks = math.ceil(inputs_embeds.shape[2] / chunk_size)

        for i in range(num_of_chunks):
            start_index = i * chunk_size
            end_index = min(start_index + chunk_size, inputs_embeds.shape[2])
            cache_size = 0 if past_key_values is None else past_key_values.get_seq_length()

            logits = self.mxq_model.infer([inputs_embeds[:, :, start_index:end_index, :]], None, cache_size)[0]

            if use_cache:
                past_key_values.update_cache_position(cache_position[start_index:end_index])

        logits = torch.tensor(logits, dtype=torch.float32, device=self.device).squeeze(0)

        if not return_dict:
            return tuple(
                v for v in [logits, past_key_values] if v is not None
            )
        return BaseModelOutputWithPast(
            last_hidden_state=logits,
            past_key_values=past_key_values,
            hidden_states=None,
            attentions=None,
        )
    
    def dispose(self):
        self.mxq_model.dispose()


class MobilintQwen2VLModel(MobilintQwen2VLPreTrainedModel, Qwen2VLModel):
    def __init__(self, config: MobilintQwen2VLConfig):
        MobilintQwen2VLPreTrainedModel.__init__(self, config)
        self.visual = MobilintQwen2VisionTransformerPretrainedModel(config.vision_config)
        self.language_model = MobilintQwen2VLTextModel(config.text_config)
        self.rope_deltas = None  # cache rope_deltas here
    
    def dispose(self):
        self.visual.dispose()
        self.language_model.dispose()


class MobilintQwen2VLForConditionalGeneration(MobilintQwen2VLPreTrainedModel, MobilintGenerationMixin, Qwen2VLForConditionalGeneration):
    _tied_weights_keys = []
    
    def __init__(self, config: MobilintQwen2VLConfig):
        MobilintQwen2VLPreTrainedModel.__init__(self, config)
        
        self.model = MobilintQwen2VLModel(config)
        # lm_head is done in self.model
        # So we just replace self.lm_head with identity module
        self.lm_head = nn.Identity()
    
    def get_mxq_model(self):
        return self.model.language_model.mxq_model
    
    def dispose(self):
        self.model.dispose()
        

AutoConfig.register("mobilint-qwen2_vl", MobilintQwen2VLConfig)
AutoTokenizer.register(MobilintQwen2VLConfig, fast_tokenizer_class=Qwen2TokenizerFast)
AutoProcessor.register(MobilintQwen2VLConfig, MobilintQwen2VLProcessor)
AutoModelForImageTextToText.register(MobilintQwen2VLConfig, MobilintQwen2VLForConditionalGeneration)

from ..utils.types import TransformersModelInfo

qwen2_vl_2b_instruct = TransformersModelInfo(
    original_model_id="Qwen/Qwen2-VL-2B-Instruct",
    model_id="mobilint/Qwen2-VL-2B-Instruct",
    download_url_base="https://dl.mobilint.com/model/transformers/vlm/Qwen2-VL-2B-Instruct/",
    file_list=[
        "chat_template.json",
        "config.json",
        "generation_config.json",
        "merges.txt",
        "model.safetensors",
        "preprocessor_config.json",
        "Qwen2-VL-2B-Instruct_text_model.mxq",
        "Qwen2-VL-2B-Instruct_vision_transformer.mxq",
        "tokenizer.json",
        "tokenizer_config.json",
        "video_preprocessor_config.json",
        "vocab.json",
    ],
)



import numpy as np
import torch
import sys
import os
from typing import Union
from urllib.parse import urlparse
import maccel
from ..utils.downloads import download_url_to_file
from .utils.types import TensorLike, ModelInfoSet
from .utils.preprocess import build_preprocess
from .utils.postprocess import build_postprocess
from .utils.results import Results


class MBLT_Engine:
    def __init__(self, model_cfg: dict, pre_cfg: dict, post_cfg: dict):
        self.model_cfg = model_cfg
        self.pre_cfg = pre_cfg
        self.post_cfg = post_cfg
        self.model = MXQ_Model(**self.model_cfg)
        self._preprocess = build_preprocess(self.pre_cfg)
        self._postprocess = build_postprocess(self.pre_cfg, self.post_cfg)
        self.device = torch.device("cpu")

    @classmethod
    def from_model_info_set(
        cls,
        model_info_set: ModelInfoSet,
        local_path: str = None,
        model_type: str = "DEFAULT",
        infer_mode: str = "global",
        product: str = "aries",
    ):
        """
        Create an instance of the model from a ModelInfoSet.
        """

        assert (
            model_type in model_info_set.__dict__.keys()
        ), f"model_type {model_type} not found. Available types: {model_info_set.__dict__.keys()}"
        assert (
            product
            in model_info_set.__dict__[model_type].value.model_cfg["url_dict"].keys()
        ), f"product {product} not found in model_type {model_type}. Available products: {model_info_set.__dict__[model_type].value.model_cfg['url_dict'].keys()}"
        assert (
            infer_mode
            in model_info_set.__dict__[model_type]
            .value.model_cfg["url_dict"][product]
            .keys()
        ), f"infer_mode {infer_mode} not found in model_type {model_type} for product {product}. Available infer modes: {model_info_set.__dict__[model_type].value.model_cfg['url_dict'][product].keys()}"

        model_cfg = model_info_set.__dict__[model_type].value.model_cfg
        model_cfg["local_path"] = local_path
        model_cfg["infer_mode"] = infer_mode
        model_cfg["product"] = product

        pre_cfg = model_info_set.__dict__[model_type].value.pre_cfg
        post_cfg = model_info_set.__dict__[model_type].value.post_cfg

        return cls(model_cfg, pre_cfg, post_cfg)

    def __call__(self, x: TensorLike):
        return self.model(x)

    def preprocess(self, x, **kwargs):
        return self._preprocess(x, **kwargs)

    def postprocess(self, x, **kwargs):
        pre_result = self._postprocess(x, **kwargs)
        return Results(self.pre_cfg, self.post_cfg, pre_result, **kwargs)

    def to(self, device: Union[str, torch.device]):
        self._preprocess.to(device)
        self._postprocess.to(device)

        if isinstance(device, str):
            self.device = torch.device(device)
        elif isinstance(device, torch.device):
            self.device = device
        else:
            raise TypeError(f"Got unexpected type for device={type(device)}.")

    def cpu(self):
        self.to(device="cpu")

    def gpu(self):
        self.to(device="cuda")

    def cuda(self, device: Union[str, int] = 0):
        if isinstance(device, int):
            device = f"cuda:{device}"
        elif isinstance(device, str):
            if not device.startswith("cuda:"):
                raise ValueError("Invalid device string. It should start with 'cuda:'.")

        if not torch.cuda.is_available():
            raise RuntimeError("CUDA is not available. Please check your environment.")
        self.to(device=device)

    def dispose(self):
        self.model.dispose()


class MXQ_Model:
    def __init__(
        self,
        url_dict,
        local_path: str = None,
        infer_mode: str = "global",
        product: str = "aries",
    ):
        self.infer_mode = infer_mode
        self.product = product
        if self.product not in url_dict.keys():  # execption handling
            url_dict[self.product] = {self.infer_mode: None}

        self.acc = maccel.Accelerator()
        # ----------------Core Allocation-------------------------
        mc = maccel.ModelConfig()
        if self.product == "aries":
            if self.infer_mode == "single":
                pass  # default is single with all cores
            elif self.infer_mode == "multi":
                mc.set_multi_core_mode(
                    [maccel.Cluster.Cluster0, maccel.Cluster.Cluster1]
                )
            elif self.infer_mode == "global4":
                mc.set_global4_core_mode(
                    [maccel.Cluster.Cluster0, maccel.Cluster.Cluster1]
                )
            elif self.infer_mode == "global" or self.infer_mode == "global8":
                mc.set_global8_core_mode()
            else:
                raise ValueError("Inappropriate inferece mode")
        elif self.product == "regulus":
            assert (
                self.infer_mode == "single"
            ), "Only single core mode is available on Regulus"
        else:
            raise ValueError("Inappropriate product")

        # -----------------Model Preparation-----------------------
        url = url_dict[self.product].get(self.infer_mode)
        if url is not None:
            parts = urlparse(url)
            filename = os.path.basename(parts.path)

            if local_path is None:  # default option
                model_dir = os.path.expanduser(
                    f"~/.mblt_model_zoo/vision/{product}/{infer_mode}"
                )
                os.makedirs(model_dir, exist_ok=True)
                cached_file = os.path.join(model_dir, filename)

            else:
                if local_path.endswith(".mxq"):
                    cached_file = local_path
                else:
                    os.makedirs(local_path, exist_ok=True)
                    cached_file = os.path.join(local_path, filename)

            if not os.path.exists(cached_file):
                sys.stderr.write(f'Downloading: "{url}" to {cached_file}\n')
                download_url_to_file(url, cached_file, progress=True)

        else:
            if local_path is not None:
                if os.path.isfile(local_path):
                    cached_file = local_path
                else:
                    raise ValueError(
                        "The model should be prepared on server or local path"
                    )
            else:
                raise ValueError("The model should be prepared on server or local path")

        # ----------------Initialize Model----------------------
        self.model = maccel.Model(cached_file, mc)
        self.model.launch(self.acc)

    def __call__(self, x: TensorLike):
        if isinstance(x, torch.Tensor):
            x = x.cpu().numpy()
        assert isinstance(x, np.ndarray), "Input should be a numpy array"

        npu_outs = self.model.infer(x)
        return npu_outs

    def dispose(self):
        """Dispose the model."""
        self.model.dispose()



from ..utils.types import ModelInfo, ModelInfoSet
from ..wrapper import MBLT_Engine


class YOLOv5su_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov5su/aries/single/yolov5su.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov5su/aries/multi/yolov5su.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov5su/aries/global/yolov5su.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [640, 640],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "nl": 3,  # Number of detection layers
        },
    )
    DEFAULT = COCO_V1


class YOLOv5s6u_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov5s6u/aries/single/yolov5s6u.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov5s6u/aries/multi/yolov5s6u.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov5s6u/aries/global/yolov5s6u.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [1280, 1280],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "nl": 4,  # Number of detection layers
        },
    )
    DEFAULT = COCO_V1


class YOLOv5mu_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov5mu/aries/single/yolov5mu.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov5mu/aries/multi/yolov5mu.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov5mu/aries/global/yolov5mu.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [640, 640],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "nl": 3,  # Number of detection layers
        },
    )
    DEFAULT = COCO_V1


class YOLOv5m6u_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov5m6u/aries/single/yolov5m6u.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov5m6u/aries/multi/yolov5m6u.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov5m6u/aries/global/yolov5m6u.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [1280, 1280],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "nl": 4,  # Number of detection layers
        },
    )
    DEFAULT = COCO_V1


class YOLOv5lu_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov5lu/aries/single/yolov5lu.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov5lu/aries/multi/yolov5lu.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov5lu/aries/global/yolov5lu.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [640, 640],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "nl": 3,  # Number of detection layers
        },
    )
    DEFAULT = COCO_V1


class YOLOv5l6_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov5l6/aries/single/yolov5l6.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov5l6/aries/multi/yolov5l6.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov5l6/aries/global/yolov5l6.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [1280, 1280],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "anchors": [
                [19, 27, 44, 40, 38, 94],  # P3/8
                [96, 68, 86, 152, 180, 137],  # P4/16
                [140, 301, 303, 264, 238, 542],  # P5/32
                [436, 615, 739, 380, 925, 792],  # P6/64
            ],
        },
    )
    DEFAULT = COCO_V1


class YOLOv5l6u_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov5l6u/aries/single/yolov5l6u.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov5l6u/aries/multi/yolov5l6u.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov5l6u/aries/global/yolov5l6u.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [1280, 1280],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "nl": 4,  # Number of detection layers
        },
    )
    DEFAULT = COCO_V1


class YOLOv5xu_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov5xu/aries/single/yolov5xu.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov5xu/aries/multi/yolov5xu.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov5xu/aries/global/yolov5xu.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [640, 640],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "nl": 3,  # Number of detection layers
        },
    )
    DEFAULT = COCO_V1


class YOLOv5x6_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov5x6/aries/single/yolov5x6.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov5x6/aries/multi/yolov5x6.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov5x6/aries/global/yolov5x6.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [1280, 1280],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "anchors": [
                [19, 27, 44, 40, 38, 94],  # P3/8
                [96, 68, 86, 152, 180, 137],  # P4/16
                [140, 301, 303, 264, 238, 542],  # P5/32
                [436, 615, 739, 380, 925, 792],  # P6/64
            ],
        },
    )
    DEFAULT = COCO_V1


class YOLOv5x6u_Set(ModelInfoSet):
    COCO_V1 = ModelInfo(
        model_cfg={
            "url_dict": {
                "aries": {
                    "single": "https://dl.mobilint.com/model/vision/object_detection/yolov5x6u/aries/single/yolov5x6u.mxq",
                    "multi": "https://dl.mobilint.com/model/vision/object_detection/yolov5x6u/aries/multi/yolov5x6u.mxq",
                    "global": "https://dl.mobilint.com/model/vision/object_detection/yolov5x6u/aries/global/yolov5x6u.mxq",
                    "global4": None,
                    "global8": None,
                },
                "regulus": {"single": None},
            },
        },
        pre_cfg={
            "Reader": {
                "style": "numpy",
            },
            "YoloPre": {
                "img_size": [1280, 1280],
            },
            "SetOrder": {"shape": "CHW"},
        },
        post_cfg={
            "task": "object_detection",
            "nc": 80,  # Number of classes
            "nl": 4,  # Number of detection layers
        },
    )
    DEFAULT = COCO_V1


def YOLOv5su(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv5su_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv5s6u(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv5s6u_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv5mu(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv5mu_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv5m6u(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv5m6u_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv5lu(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv5lu_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv5l6(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv5l6_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv5l6u(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv5l6u_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv5xu(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv5xu_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv5x6(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv5x6_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )


def YOLOv5x6u(
    local_path: str = None,
    model_type: str = "DEFAULT",
    infer_mode: str = "global",
    product: str = "aries",
) -> MBLT_Engine:
    return MBLT_Engine.from_model_info_set(
        YOLOv5x6u_Set,
        local_path=local_path,
        model_type=model_type,
        infer_mode=infer_mode,
        product=product,
    )



from typing import Optional, Union

import maccel
import torch

from transformers import (
    Cache,
    BertPreTrainedModel,
    BertForMaskedLM,
    BertConfig,
    BertTokenizerFast,
    AutoConfig,
    AutoModel,
    AutoTokenizer,
    AutoModelForMaskedLM,
)
from transformers.models.bert.modeling_bert import BertEmbeddings, BertPooler, BertOnlyMLMHead
from transformers.modeling_outputs import BaseModelOutputWithPoolingAndCrossAttentions
from transformers.utils import logging


logger = logging.get_logger(__name__)


class MobilintBertConfig(BertConfig):
    model_type = "mobilint-bert"

    def __init__(
        self,
        mxq_path: str = "",
        dev_no: int = 0,
        **kwargs,
    ):
        self.mxq_path = mxq_path
        self.dev_no = dev_no

        super().__init__(**kwargs)


class MobilintBertPreTrainedModel(BertPreTrainedModel):
    config: MobilintBertConfig
    load_tf_weights = None
    supports_gradient_checkpointing = False
    _supports_sdpa = False


class MobilintBertModel(MobilintBertPreTrainedModel):
    _no_split_modules = ["BertEmbeddings", "BertLayer"]

    def __init__(self, config: MobilintBertConfig, add_pooling_layer=True):
        super().__init__(config)
        self.config = config

        self.embeddings = BertEmbeddings(config)

        self.pooler = BertPooler(config) if add_pooling_layer else None

        self.attn_implementation = config._attn_implementation
        self.position_embedding_type = config.position_embedding_type

        self.dev_no = config.dev_no
        self.acc = maccel.Accelerator(self.dev_no)
        mc = maccel.ModelConfig()
        mc.set_single_core_mode(1)
        self.mxq_model = maccel.Model(f"{config.name_or_path}/{config.mxq_path}", mc)
        self.mxq_model.launch(self.acc)
        self.mxq_model.reset_cache_memory()

    def get_input_embeddings(self):
        return self.embeddings.word_embeddings

    def set_input_embeddings(self, value):
        self.embeddings.word_embeddings = value

    def _prune_heads(self, heads_to_prune):
        raise NotImplementedError("_prune_heads is not available since self.encoder is implemented in mxq")

    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        past_key_values: Optional[list[torch.FloatTensor]] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        cache_position: Optional[torch.Tensor] = None,
    ) -> Union[tuple[torch.Tensor], BaseModelOutputWithPoolingAndCrossAttentions]:
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if self.config.is_decoder:
            use_cache = use_cache if use_cache is not None else self.config.use_cache
        else:
            use_cache = False

        if head_mask is not None:
            logger.warning_once("head_mask is not supported.")

        if encoder_hidden_states is not None:
            logger.warning_once("encoder_hidden_states is not supported.")
            
        if encoder_attention_mask is not None:
            logger.warning_once("encoder_attention_mask is not supported.")
        
        if use_cache:
            logger.warning_once("use_cache is not supported.")

        if output_attentions:
            logger.warning_once("output_attentions is not supported.")

        if output_hidden_states:
            logger.warning_once("output_hidden_states is not supported.")
        
        if cache_position is not None:
            logger.warning_once("cache_position is not supported.")

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            self.warn_if_padding_and_no_attention_mask(input_ids, attention_mask)
            input_shape = input_ids.size()
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        batch_size, seq_length = input_shape
        device = input_ids.device if input_ids is not None else inputs_embeds.device

        past_key_values_length = 0
        if past_key_values is not None:
            past_key_values_length = (
                past_key_values[0][0].shape[-2]
                if not isinstance(past_key_values, Cache)
                else past_key_values.get_seq_length()
            )

        if token_type_ids is None:
            if hasattr(self.embeddings, "token_type_ids"):
                buffered_token_type_ids = self.embeddings.token_type_ids[:, :seq_length]
                buffered_token_type_ids_expanded = buffered_token_type_ids.expand(batch_size, seq_length)
                token_type_ids = buffered_token_type_ids_expanded
            else:
                token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=device)

        embedding_output = self.embeddings(
            input_ids=input_ids,
            position_ids=position_ids,
            token_type_ids=token_type_ids,
            inputs_embeds=inputs_embeds,
            past_key_values_length=past_key_values_length,
        )
        
        embedding_output = embedding_output.type(torch.float32).cpu().numpy()

        sequence_output = self.mxq_model.infer([embedding_output])[0]
        sequence_output = torch.tensor(sequence_output, dtype=torch.float32, device=self.device)
        pooled_output = self.pooler(sequence_output) if self.pooler is not None else None

        if not return_dict:
            return (sequence_output, pooled_output) + (past_key_values,)

        return BaseModelOutputWithPoolingAndCrossAttentions(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            past_key_values=past_key_values,
            hidden_states=None,
            attentions=None,
            cross_attentions=None,
        )

    def dispose(self):
        self.mxq_model.dispose()


class MobilintBertForMaskedLM(MobilintBertPreTrainedModel, BertForMaskedLM):
    _tied_weights_keys = ["predictions.decoder.bias", "cls.predictions.decoder.weight"]

    def __init__(self, config):
        MobilintBertPreTrainedModel.__init__(self, config)

        if config.is_decoder:
            logger.warning(
                "If you want to use `BertForMaskedLM` make sure `config.is_decoder=False` for "
                "bi-directional self-attention."
            )

        self.bert = MobilintBertModel(config, add_pooling_layer=False)
        self.cls = BertOnlyMLMHead(config)

        # Initialize weights and apply final processing
        self.post_init()
    
    def dispose(self):
        self.bert.dispose()


AutoConfig.register("mobilint-bert", MobilintBertConfig)
AutoModel.register(MobilintBertConfig, MobilintBertForMaskedLM)
AutoTokenizer.register(MobilintBertConfig, fast_tokenizer_class=BertTokenizerFast)
AutoModelForMaskedLM.register(MobilintBertConfig, MobilintBertForMaskedLM)

from ..utils.types import TransformersModelInfo

bert_base_uncased = TransformersModelInfo(
    original_model_id="google-bert/bert-base-uncased",
    model_id="mobilint/bert-base-uncased",
    download_url_base="https://dl.mobilint.com/model/transformers/llm/bert-base-uncased/",
    file_list=[
        "bert-base-uncased.mxq",
        "config.json",
        "model.safetensors",
        "tokenizer.json",
        "tokenizer_config.json",
        "vocab.txt",
    ],
)
